class PupaerUT extends Pupae;

//Mesh reimported to fix the green foot on the model, otherwise it's the same
#exec MESH IMPORT MESH=Pupae2 ANIVFILE=Models\Pupae1_a.3d DATAFILE=Models\Pupae1_d.3d X=0 Y=0 Z=0
#exec MESH ORIGIN MESH=Pupae2 X=0 Y=0 Z=-90 YAW=64 PITCH=0 ROLL=-64

#exec MESH SEQUENCE MESH=Pupae2 SEQ=All		STARTFRAME=0	 NUMFRAMES=171
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Bite     STARTFRAME=0    NUMFRAMES=15  RATE=15  Group=Attack
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Crawl    STARTFRAME=15   NUMFRAMES=20  RATE=20
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Dead     STARTFRAME=35   NUMFRAMES=18  RATE=15
#exec MESH SEQUENCE MESH=Pupae2 SEQ=TakeHit  STARTFRAME=36   NUMFRAMES=1
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Fighter  STARTFRAME=53   NUMFRAMES=6   RATE=6
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Lunge    STARTFRAME=59   NUMFRAMES=15  RATE=15  Group=Attack
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Munch    STARTFRAME=74   NUMFRAMES=8   RATE=15
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Pick     STARTFRAME=82   NUMFRAMES=10  RATE=15
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Stab     STARTFRAME=92   NUMFRAMES=10  RATE=15  Group=Attack
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Tear     STARTFRAME=102  NUMFRAMES=28  RATE=15
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Dead2    STARTFRAME=130  NUMFRAMES=18  RATE=15
#exec MESH SEQUENCE MESH=Pupae2 SEQ=Dead3    STARTFRAME=148  NUMFRAMES=23  RATE=15

//Texture reimported in case it needs cleaned up or altered, for now it's identical to default
#exec texture IMPORT NAME=JPupae2Tex FILE=Textures\JPupae2Tex.bmp GROUP=Skins DETAIL=Dirty
#exec MESHMAP scale MESHMAP=Pupae2 X=0.05 Y=0.05 Z=0.1
#exec MESHMAP SETTEXTURE MESHMAP=Pupae2 NUM=1 TEXTURE=JPupae2Tex

#exec MESH NOTIFY MESH=Pupae2 SEQ=Dead TIME=0.52 FUNCTION=LandThump

var(Jumping) int JumpVelocity;  		//Lunge power/distance
var(Jumping) bool CanLeapUp;  		//Allows the pupae to leap around much more
var(Jumping) int LeapUpStrength;  	//Leap power/distance
var(Jumping) int MeleeRangeCheck; 	//Default check for 'is in melee range'

function bool InMeleeRange(Pawn Other)
{
	local vector Dist2d;

	Dist2d = Location - Enemy.Location;
	Dist2d.Z = 0;
	return Abs(Location.Z - Other.Location.Z) <= MeleeRangeCheck;
}

state Charging
{
	ignores SeePlayer, HearNoise;
	
	function HitWall(vector HitNormal, actor Wall)
	{
		if (Physics == PHYS_Falling)
			return;
		if ( Wall.IsA('Mover') && Mover(Wall).HandleDoor(self) )
		{
			bSpecialGoal = true;
			if ( SpecialPause > 0 )
				Acceleration = vect(0,0,0);
			GotoState('Charging', 'SpecialNavig');
			return;
		}
		Focus = Destination;

		if(CanLeapUp)
		{
			Velocity = LeapUpStrength * Normal(Enemy.Location + Enemy.CollisionHeight * vect(0,0,0.90) - Location);
			SetPhysics(PHYS_Falling);
		}

		if (PickWallAdjust())
		{
			if ( Physics == PHYS_Falling )
				SetFall();
			else
				gotoState('Charging', 'AdjustFromWall');
		}
		else
			MoveTimer = -1.0;
	}

	function Timer()
	{
		bReadyToAttack = True;
		if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		{
			GotoState('Attacking');
			return;
		}
		Target = Enemy;

		if( InMeleeRange(Enemy))
			GotoState('MeleeAttack');
		else if ( bHasRangedAttack && (FRand() > 0.7 + 0.1 * skill) )
			GotoState('RangedAttack');
		else if ( bHasRangedAttack && !bMovingRangedAttack)
		{
			if ( FRand() < CombatStyle * 0.8 ) //then keep charging
				SetTimer(1.0,false);
			else
				GotoState('Attacking');
		}
	}

CloseIn:
	if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		GotoState('Attacking');

	if ( Enemy.Region.Zone.bWaterZone )
	{
		if (!bCanSwim)
			GotoState('TacticalMove', 'NoCharge');
	}
	else if (!bCanFly && !bCanWalk)
		GotoState('TacticalMove', 'NoCharge');

	if (Physics == PHYS_Falling)
	{
		DesiredRotation = Rotator(Enemy.Location - Location);
		Focus = Enemy.Location;
		Destination = Enemy.Location;
		WaitForLanding();
	}
	if ( (Intelligence <= BRAINS_Reptile) || actorReachable(Enemy) )
	{
		bCanFire = true;
		if ( FRand() < 0.3 )
			PlayThreateningSound();
		MoveToward(Enemy);
		if (bFromWall)
		{
			bFromWall = false;
			if (PickWallAdjust())
			{
				if ( Physics == PHYS_Falling )
					SetFall();
				else
					StrafeFacing(Destination, Enemy);
			}
			else
				GotoState('TacticalMove', 'NoCharge');
		}
	}
	else
	{
NoReach:
		if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
			GotoState('Attacking');
		bCanFire = false;
		bFromWall = false;
		log("route to enemy "$Enemy);
		if (!FindBestPathToward(Enemy))
		{
			Sleep(0.0);
			GotoState('TacticalMove', 'NoCharge');
		}
SpecialNavig:
		if ( SpecialPause > 0.0 )
		{
			Target = Enemy;
			bFiringPaused = true;
			NextState = 'Charging';
			NextLabel = 'Moving';
			GotoState('RangedAttack');
		}
Moving:
		if ( !IsAnimating() )
			AnimEnd();
		if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
			gotoState('Attacking');
		if (MoveTarget!=none)
		{
			if( InMeleeRange(Enemy))
			{
				bCanFire = true;
				StrafeFacing(MoveTarget.Location, Enemy);
			}
			else
			{
				if ( !bCanStrafe || !LineOfSightTo(Enemy) ||
						(Skill - 2 * FRand() + (Normal(Enemy.Location - Location - vect(0,0,1) * (Enemy.Location.Z - Location.Z))
												Dot Normal(MoveTarget.Location - Location - vect(0,0,1) * (MoveTarget.Location.Z - Location.Z))) < 0) )
				{
					if ( GetAnimGroup(AnimSequence) == 'MovingAttack' )
					{
						AnimSequence = '';
						TweenToRunning(0.12);
					}
					MoveToward(MoveTarget);
				}
				else
				{
					bCanFire = true;
					StrafeFacing(MoveTarget.Location, Enemy);
				}
				if ( !bFromWall && (FRand() < 0.5) )
					PlayThreateningSound();
			}
		}
	}
	if( InMeleeRange(Enemy))
		goto('GotThere');
	if ( bIsPlayer || (!bFromWall && bHasRangedAttack && (FRand() > CombatStyle + 0.1)) )
		gotoState('Attacking');
	MoveTimer = 0.0;
	bFromWall = false;
	goto('CloseIn');

GotThere:
	if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		gotoState('Attacking');
	Target = Enemy;
	GotoState('MeleeAttack');

TakeHit:
	TweenToRunning(0.12);
	if (MoveTarget == Enemy)
	{
		bCanFire = true;
		MoveToward(MoveTarget);
	}
	Goto('Charge');
}

function PlayMeleeAttack()
{
	local float dist, decision;

	if ( Target == None )
	{
		PlayWaiting();
		return;
	}
	decision = FRand();
	dist = VSize(Target.Location - Location);
	if (dist > CollisionRadius + Target.CollisionRadius + 45)
		decision = 0.0;

	if (Physics == PHYS_Falling)
		decision = 1.0;
	if (Target == None)
		decision = 1.0;

	if (decision < 0.15)
	{
		PlaySound(Lunge, SLOT_Interact);
		Enable('Bump');
		PlayAnim('Lunge');
		Velocity = JumpVelocity * Normal(Target.Location + Target.CollisionHeight * vect(0,0,0.75) - Location);
		if (dist > CollisionRadius + Target.CollisionRadius + 35)
			Velocity.Z += 0.7 * dist;
		SetPhysics(PHYS_Falling);
	}
	else
	{
		PlaySound(Stab, SLOT_Interact);
		PlayAnim('Stab');
		MeleeRange = 50;
		MeleeDamageTarget(BiteDamage, vect(0,0,0));
		MeleeRange = Default.MeleeRange;
	}
}

state TacticalMove
{
	ignores SeePlayer, HearNoise;

	function HitWall(vector HitNormal, actor Wall)
	{
		if (Physics == PHYS_Falling)
			return;
			
		if(CanLeapUp)
		{
			Velocity = LeapUpStrength * Normal(Enemy.Location + Enemy.CollisionHeight * vect(0,0,0.90) - Location);
			SetPhysics(PHYS_Falling);
		}

		Focus = Destination;
		if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		{
			GotoState('Attacking');
			return;
		}
		if ( bChangeDir || (FRand() < 0.5) || (((Enemy.Location - Location) Dot HitNormal) < 0) )
		{
			DesiredRotation = Rotator(Enemy.Location - location);
			GiveUpTactical(false);
		}
		else
		{
			bChangeDir = true;
			Destination = Location - HitNormal * FRand() * 500;
		}
	}
}

state RangedAttack
{
	ignores SeePlayer, HearNoise, Bump;

FaceTarget:
	Disable('AnimEnd');
	Sleep(0.0);
	if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		GotoState('Attacking');
	Target = Enemy;
	if (NeedToTurn(Enemy.Location))
	{
		PlayTurning();
		TurnToward(Enemy);
		TweenToFighter(0.1);
	}
	FinishAnim();

	if (Enemy != none && InMeleeRange(Enemy))
		GotoState('MeleeAttack', 'ReadyToAttack');
}

state MeleeAttack
{
	ignores SeePlayer, HearNoise;

	singular function Bump(actor Other)
	{
		Disable('Bump');
		if ( (Other == Target) && (AnimSequence == 'Lunge') )
			if (MeleeDamageTarget(LungeDamage, vect(0,0,0)))
			{
				if (FRand() < 0.5)
					PlaySound(Tear, SLOT_Interact);
				else
					PlaySound(Bite, SLOT_Interact);
			}
	}

	function KeepAttacking()
	{
		if ((Enemy == None) || (Enemy == self )|| (Enemy.Health <= 0) || (Enemy.bDeleteme) || (InMeleeRange(Enemy)))
			GotoState('Attacking');
				else Velocity = JumpVelocity * Normal(Enemy.Location + Enemy.CollisionHeight * vect(0,0,0.75) - Location);
	}

	function EnemyNotVisible()
	{
		GotoState('Attacking');
	}

	function AnimEnd()
	{
		GotoState('MeleeAttack', 'DoneAttacking');
	}

	function BeginState()
	{
		if ( Enemy != none && Enemy.health > 0 && !Enemy.bdeleteme && Enemy != self )
			Target = Enemy;
		Disable('AnimEnd');
		bReadyToAttack = false;
	}
	
	function MediumAttackRangeCheck()
	{
		if ( VSize(Location - Enemy.Location) < 300 )
			Velocity = JumpVelocity * Normal(Enemy.Location + Enemy.CollisionHeight * vect(0,0,0.75) - Location);
	}

Begin:
	if ( Enemy != None )
		DesiredRotation = Rotator(Enemy.Location - Location);
	if ( skill < 3 )
		TweenToFighter(0.15);
	else
		TweenToFighter(0.11);

FaceTarget:
	if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		GotoState('Attacking');
	Target = Enemy;
	Disable('AnimEnd');
	Sleep(0.0);
	Acceleration = vect(0,0,0); //stop
	if (NeedToTurn(Enemy.Location))
	{
		PlayTurning();
		TurnToward(Enemy);
		TweenToFighter(0.1);
	}
	FinishAnim();
	OldAnimRate = 0;	// force no tween

	if( !InMeleeRange(Enemy))
	{
		MediumAttackRangeCheck();
		PlayMeleeAttack();
		FinishAnim();
	}

ReadyToAttack:
	if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		GotoState('Attacking');
	Target = Enemy;
	DesiredRotation = Rotator(Enemy.Location - Location);
	PlayMeleeAttack();
	FinishAnim();
	Enable('AnimEnd');
	Sleep(0.0);
Attacking:
	if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		GotoState('Attacking');
	if (NeedToTurn(Enemy.Location))
	{
		TurnToward(Enemy);
		Goto('Attacking');
	}
DoneAttacking:
	Disable('AnimEnd');
	Sleep(0.0);
	KeepAttacking();
	if ( FRand() < 0.3 - 0.1 * skill )
	{
		Acceleration = vect(0,0,0); //stop
		DesiredRotation = Rotator(Enemy.Location - Location);
		PlayChallenge();
		FinishAnim();
		TweenToFighter(0.1);
	}
	Goto('FaceTarget');
}

state Attacking
{
	ignores SeePlayer, HearNoise, Bump, HitWall;

	function ChooseAttackMode()
	{
		local eAttitude AttitudeToEnemy;
		local pawn changeEn;

		if ( Enemy == none || Enemy.health <= 0 || Enemy.bdeleteme || Enemy == self )
		{
			if (Orders == 'Attacking')
				Orders = '';
			WhatToDoNext('','');
			return;
		}

		if ( (AlarmTag != '') && Enemy.bIsPlayer )
		{
			if (AttitudeToPlayer > ATTITUDE_Ignore)
			{
				GotoState('AlarmPaused', 'WaitForPlayer');
				return;
			}
			else if ( (AttitudeToPlayer != ATTITUDE_Fear) || bInitialFear )
			{
				GotoState('TriggerAlarm');
				return;
			}
		}

		AttitudeToEnemy = AttitudeTo(Enemy);

		if (AttitudeToEnemy == ATTITUDE_Fear)
		{
			GotoState('Retreating');
			return;
		}

		else if (AttitudeToEnemy == ATTITUDE_Threaten)
		{
			GotoState('Threatening');
			return;
		}

		else if (AttitudeToEnemy == ATTITUDE_Friendly)
		{
			if (Enemy.bIsPlayer)
				GotoState('Greeting');
			else
				WhatToDoNext('','');
			return;
		}

		else if (!LineOfSightTo(Enemy))
		{
			if ( (OldEnemy != none  )
					&& (AttitudeTo(OldEnemy) == ATTITUDE_Hate) && LineOfSightTo(OldEnemy) && !OldEnemy.bDeleteme)
			{
				changeEn = enemy;
				enemy = oldenemy;
				oldenemy = changeEn;
			}
			else
			{
				if ( (Orders == 'Guarding') && !LineOfSightTo(OrderObject) )
					GotoState('Guarding');
				else if ( !bHasRangedAttack || VSize(Enemy.Location-Location)>(600+(FRand()*RelativeStrength(Enemy)-CombatStyle)*600) )
					GotoState('Hunting');
				else if ( bIsBoss || (Intelligence > BRAINS_None) )
				{
					HuntStartTime = Level.TimeSeconds;
					NumHuntPaths = 0;
					GotoState('StakeOut');
				}
				else WhatToDoNext('Waiting', 'TurnFromWall');
				return;
			}
		}

		else if ( (TeamLeader != None) && TeamLeader.ChooseTeamAttackFor(self) )
			return;

		if (bReadyToAttack )
		{
			Target = Enemy;
			if( InMeleeRange(Enemy))
			{
				GotoState('MeleeAttack');
				return;
			}
			else if (bMovingRangedAttack)
				SetTimer(TimeBetweenAttacks, False);
			else if (bHasRangedAttack && (bIsPlayer || enemy.bIsPlayer) && CanFireAtEnemy() )
			{
				if (!bIsPlayer || (2.5 * FRand() > Skill) )
				{
					GotoState('RangedAttack');
					return;
				}
			}
		}

		if ( !bHasRangedAttack )
			GotoState('Charging');
		else
			GotoState('TacticalMove');
	}

	function EnemyNotVisible()
	{
	}

	function Timer()
	{
		bReadyToAttack = True;
	}

	function BeginState()
	{
		if ( TimerRate <= 0.0 )
			SetTimer(TimeBetweenAttacks  * (1.0 + FRand()),false);
		if (Physics == PHYS_None)
			SetMovementPhysics();
	}

Begin:
	ChooseAttackMode();
}

DefaultProperties
{
	MeleeRangeCheck=80
	CanLeapUp=True
	LeapUpStrength=600
	Mesh=LodMesh'Pupae2'
	JumpVelocity=450
	Intelligence=BRAINS_REPTILE
	MeleeRange=280.000000
}
